/*
# Copyright 2022-2023 Advanced Micro Devices Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#
*/
const char* xilinx_vart_version() {
    return "vart-b7953a2a9f60e23efdfced5c186328dd1449665c"; 
}

static char vart_version[] = "Xilinx vart-mem-manager Version: 3.5.0-b7953a2a9f60e23efdfced5c186328dd1449665c  2023-09-12-23:21:48 ";
const char* xilinx_version() {
  return vart_version;
}

