/*
 * Copyright 2022-2023 Advanced Micro Devices Inc.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <UniLog/ErrorCode.hpp>

REGISTER_ERROR_CODE(VART_RUNNER_CONSTRUCTION_FAIL, "Cannot create runner", "");

REGISTER_ERROR_CODE(VART_TENSOR_BUFFER_CONSTRUCTION_FAIL,
                    "Cannot create tensor buffer", "");
REGISTER_ERROR_CODE(VART_TENSOR_BUFFER_INVALID_INDEX, "Invalid index!", "");
REGISTER_ERROR_CODE(VART_TENSOR_BUFFER_UNSUPPORT_FORMAT, "unsupported format!",
                    "");
REGISTER_ERROR_CODE(VART_FAILED_FILE_OPERATION, "failed file operation!", "");
REGISTER_ERROR_CODE(VART_UNEXPECTED_FILE_SIZE, "Unexpected file size!", "");
