# -*- coding: UTF-8 -*-

# Copyright 2021 Xilinx Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

VAITRACE_VER = "3.5.0"
VAITRACE_BUILD_DATE = "2023-09-12-23:21:46"
VAITRACE_GIT_VER = "b7953a2a9"
TOP_PROJECT = "vart"

VAITRACE_GIT_INTERNAL_PREV_VER="671d17e"
