/*
 * Copyright 2022-2023 Advanced Micro Devices Inc.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
#include "vitis/ai/time_measure.hpp"
#include <iostream>
using namespace vitis::ai;

static thread_local TimeMeasure dpu_time_measure_;

TimeMeasure::TimeMeasure() : consuming_time_{0} {}
TimeMeasure::~TimeMeasure() {}

TimeMeasure& TimeMeasure::getThreadLocalForDpu() { return dpu_time_measure_; }

void TimeMeasure::reset() { consuming_time_ = 0; }

void TimeMeasure::add(int time) { consuming_time_ += time; }

int TimeMeasure::get() { return consuming_time_; }
